'''
General metric utilities and constants.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from .format import BoundaryFormat


METRIC_DEFAULTS = {
    'boundary_format' : BoundaryFormat.mass,
    'permuted' : False,
    'one_minus' : False,
    'return_parts' : False
}



'''
Window-based segmentation evaluation metrics package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from decimal import Decimal
from ..metric import METRIC_DEFAULTS
from ..format import BoundaryFormat, convert_positions_to_masses
from ..util.math import mean


WINDOW_METRIC_DEFAULTS = dict(METRIC_DEFAULTS)
WINDOW_METRIC_DEFAULTS.update({
    'window_size' : None,
    'fnc_round' : round,
    'permuted' : True
})


def __compute_window_size__(reference, fnc_round, boundary_format):
    '''
    Compute a window size from a dict of segment masses.
    
    :param masses: A dict of segment masses.
    :type masses: dict
    '''
    all_masses = list()
    # Define fnc
    def __list_coder_masses__(inner_coder_masses):
        '''
        Recursively collect all masses.
        
        :param inner_coder_masses: Either a dict of dicts, or dict of a list of
            masses.
        :type inner_coder_masses: dict or list
        '''
        if isinstance(inner_coder_masses, list) or \
            isinstance(inner_coder_masses, tuple):
            all_masses.extend(inner_coder_masses)
        elif isinstance(inner_coder_masses, dict):
            for cur_inner_coder_masses in inner_coder_masses.values():
                __list_coder_masses__(cur_inner_coder_masses)
    if boundary_format == BoundaryFormat.position:
        reference = convert_positions_to_masses(reference)
    # Recurse and list all masses
    __list_coder_masses__(reference)
    # Convert to floats
    all_masses = [Decimal(mass) for mass in all_masses]
    # Calculate
    avg = mean(all_masses) / Decimal('2')
    window_size = int(fnc_round(avg))
    return window_size if window_size > 1 else 2


def compute_window_size(reference, **kwargs):
    metric_kwargs = dict(WINDOW_METRIC_DEFAULTS)
    metric_kwargs.update(kwargs)
    del metric_kwargs['one_minus']
    del metric_kwargs['permuted']
    del metric_kwargs['window_size']
    del metric_kwargs['return_parts']
    return __compute_window_size__(reference, **metric_kwargs)

    


'''
Tests the segmentation versions of Scott's and Fleiss' Pi.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from .pi import fleiss_pi_linear
from ..data.samples import (KAZANTSEVA2012_G5, KAZANTSEVA2012_G2, 
    COMPLETE_AGREEMENT, LARGE_DISAGREEMENT)


class TestPi(unittest.TestCase):
    '''
    Test segmentation versions of Scott's Pi and Fleiss' Multi-Pi.
    '''
    # pylint: disable=R0904

    def test_fliess_pi_g5(self):
        '''
        Test Pi upon Group 5 of Kazantseva (2012) data.
        '''
        self.assertEqual(fleiss_pi_linear(KAZANTSEVA2012_G5),
                         Decimal('0.2307643892167668335086819753'))

    def test_fliess_pi_g5_ch1(self):
        '''
        Test Pi upon Group 5, Chapter 1, of Kazantseva (2012) data.
        '''
        data = {'ch1' : KAZANTSEVA2012_G5['ch1']}
        self.assertEqual(fleiss_pi_linear(data),
                         Decimal('0.1906323185011709601873536298'))

    def test_fleiss_pi_g2(self):
        '''
        Test Pi upon Group 2 of Kazantseva (2012) data.
        '''
        self.assertEqual(fleiss_pi_linear(KAZANTSEVA2012_G2),
                         Decimal('0.4018239928733601859131343866'))

    def test_fleiss_pi_g2_ch2(self):
        '''
        Test Pi upon Group 2, Chapter 2, of Kazantseva (2012) data.
        '''
        data = {'ch2' : KAZANTSEVA2012_G2['ch2']}
        # Test
        self.assertEqual(fleiss_pi_linear(data),
                         Decimal('0.5192587209302325581395348837'))

    def test_fleiss_pi_disagree(self):
        '''
        Test Pi upon a hypothetical dataset containing large disagreement.
        '''
        data = LARGE_DISAGREEMENT
        self.assertEqual(fleiss_pi_linear(data),
                         Decimal('-0.5757942099675148626179719687'))

    def test_fleiss_pi(self):
        '''
        Test Fleiss' Pi.
        '''
        # pylint: disable=C0324,C0103
        data1 = {'i1': {'c1' : [2,8,2,1],
                        'c2' : [2,1,7,2,1]}}
        pi1  = fleiss_pi_linear(data1)
        pi1f = fleiss_pi_linear(data1)
        self.assertEqual(pi1,
                         Decimal('0.7090909090909090909090909091'))
        self.assertEqual(pi1,pi1f)
        data2 = {'i1': {'c1' : [2, 8, 2, 1],
                        'c2' : [11, 2]}}
        pi2  = fleiss_pi_linear(data2)
        pi2f = fleiss_pi_linear(data2)
        self.assertEqual(pi2,
                         Decimal('0.1111111111111111111111111111'))
        self.assertEqual(pi2,pi2f)
        self.assertTrue(pi2 < pi1)
    
    def test_fleiss_pi_complete(self):
        '''
        Test Pi upon a hypothetical dataset containing complete agreement.
        '''
        # pylint: disable=C0103
        data_complete = COMPLETE_AGREEMENT
        pi = fleiss_pi_linear(data_complete)
        self.assertEqual(pi, 1.0)
        
    def test_exception_coders(self):
        '''
        Test exception.
        '''
        data = {'i1' : {'c1' : [2, 8, 2, 1]}}
        self.assertRaises(Exception, fleiss_pi_linear, data)

    def test_exception_items(self):
        '''
        Test exception.
        '''
        data = {'i1' : {'c1' : [2, 8, 2, 1],
                        'c2' : [2, 8, 2, 1]},
                'i2' : {'c1' : [2, 1, 7, 2, 1],
                        'c2' : [2, 8, 2, 1],
                        'c3' : [2, 8, 2, 1]}}
        self.assertRaises(Exception, fleiss_pi_linear, data)





'''
Test utilities, and tests for said utilities.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal

class TestCase(unittest.TestCase):
    '''
    Test WindowDiff.
    '''
    # pylint: disable=R0904,C0324
    DECIMAL_PLACES = 4

    def assertAlmostEquals(self, first, second, places=DECIMAL_PLACES,
                           msg=None, delta=None):
        '''
        Automatically converts values to floats.
        '''
        # pylint: disable=C0103,R0913
        if isinstance(first, dict) and isinstance(second, dict):
            for item in set(first.keys() + second.keys()):
                if not msg:
                    msg = '{0} != {1}'.format(first, second)
                if item not in first:
                    raise Exception(
                            '{0} not in {1}; expected {2}'.format(item, first,
                                                                  second))
                if item not in second:
                    raise Exception(
                            '{0} not in {1}; expected {2}'.format(item, second,
                                                                  first))
                self.assertAlmostEquals(first[item], second[item], places, msg,
                                        delta)
        elif (isinstance(first, list) or isinstance(first, tuple)) and \
            (isinstance(second, list) or isinstance(second, tuple)):
            if len(first) != len(second):
                raise Exception(
                        'Size mismatch; {0} != {1}'.format(first, second))
            for item in zip(first, second):
                if not msg:
                    msg = '{0} != {1}'.format(first, second)
                self.assertAlmostEquals(item[0], item[1], places, msg, delta)
        elif type(first) != type(second):
            if type(first) != float and type(first) != Decimal and \
            type(second) != float and type(second) != Decimal:
                raise Exception('Type mismatch; {0} != {1}'.format(
                                    type(first), type(second)))
        else:
            return unittest.TestCase.assertAlmostEquals(self,
                                                        float(first),
                                                        float(second),
                                                        places=places,
                                                        msg=msg,
                                                        delta=delta)


class TestTestCase(TestCase):
    '''
    Test utility tests.
    '''
    # pylint: disable=R0904
    
    def test_almost_equal_values(self):
        '''
        Test a type mistmatch.
        '''
        self.assertRaises(AssertionError, self.assertAlmostEquals,
                          {'a' : 1},
                          {'a' : 2})
    
    def test_mismatch(self):
        '''
        Test a type mistmatch.
        '''
        self.assertRaises(Exception, self.assertAlmostEquals, (), {})
    
    def test_equal_types(self):
        '''
        Test a type mistmatch.
        '''
        self.assertAlmostEquals({}, {})
        self.assertAlmostEquals([], [])
        self.assertAlmostEquals((), ())
        self.assertAlmostEquals((), [])
        self.assertAlmostEquals([], ())
        self.assertAlmostEquals(0.0, 0.0)
        self.assertAlmostEquals(0, 0)
        self.assertAlmostEquals(Decimal('0'), Decimal('0'))
        self.assertAlmostEquals(0.0, 0)
        self.assertAlmostEquals(0, 0.0)
        self.assertAlmostEquals(0, Decimal('0'))
        self.assertAlmostEquals(Decimal('0'), 0)
        self.assertAlmostEquals(0.0, Decimal('0'))
        self.assertAlmostEquals(Decimal('0'), 0.0)
    
    def test_tuple_first(self):
        self.assertRaises(Exception, self.assertAlmostEquals, (), (1))
    
    def test_tuple_second(self):
        self.assertRaises(Exception, self.assertAlmostEquals, (1), ())
    
    def test_list_first(self):
        self.assertRaises(Exception, self.assertAlmostEquals, [], [1])
    
    def test_list_second(self):
        self.assertRaises(Exception, self.assertAlmostEquals, [1], [])
    
    def test_dict_first(self):
        self.assertRaises(Exception, self.assertAlmostEquals, {}, {'a' : 1})
    
    def test_dict_second(self):
        self.assertRaises(Exception, self.assertAlmostEquals, {'a' : 1}, {})
    
        
        
        
        
        


'''
Utility functions and classes for the package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from ..compute import compute_pairwise_values


class SegmentationMetricError(Exception):
    '''
    Indicates that a runtime check has failed, and the algorithm is performing
    incorrectly, or input validation has failed.  Generation of this exception
    is tested.
        
    :param message: Explanation for the exception.
    :type message: str
    '''
    
    def __init__(self, message):
        '''
        Initializer.
        
        :param message: Explanation for the exception.
        :type message: str
        '''
        Exception.__init__(self, message)


def __fnc_metric__(fnc_metric, args, kwargs, kw_defaults):
    # pylint: disable=W0142
    from ..data import Dataset
    # Create default keyword arguments
    metric_kwargs = dict(kw_defaults)
    metric_kwargs.update(kwargs)
    # Initialize arguments
    hypothesis = None
    reference = None
    dataset = None
    # Parse arguments
    if len(args) == 2:
        hypothesis, reference = args
    elif len(args) == 1:
        dataset = args[0]
    elif len(args) == 0:
        if 'hypothesis' in kwargs and 'reference' in kwargs:
            hypothesis = kwargs['hypothesis']
            reference = kwargs['reference']
            del metric_kwargs['hypothesis']
            del metric_kwargs['reference']
        elif 'dataset' in kwargs:
            dataset = kwargs['dataset']
            del metric_kwargs['dataset']
    # Compute
    if dataset:
        # Compute pairwise values over all coders in a dataset
        metric_kwargs['boundary_format'] = dataset.boundary_format
        return compute_pairwise_values(fnc_metric, dataset, **metric_kwargs)
    elif hypothesis and reference:
        # Compute values between hypotheses (i.e, automatic) and reference 
        # (i.e., manual) coder segmentations
        if isinstance(hypothesis, Dataset) and isinstance(reference, Dataset):
            # Compare pairwise values between coders paired from two datasets
            metric_kwargs['boundary_format'] = hypothesis.boundary_format
            if hypothesis.boundary_format is not reference.boundary_format:
                raise SegmentationMetricError(
                    'Datasets contain differing boundary formats; {0} != {1}'\
                        .format(hypothesis.boundary_format, reference.boundary_format))
            return compute_pairwise_values(fnc_metric, hypothesis, reference, **metric_kwargs)
        else:
            # Compare a single pair of segmentations
            del metric_kwargs['permuted']
            return fnc_metric(hypothesis, reference, **metric_kwargs)
    # Except if insufficient arguments supplied
    raise SegmentationMetricError('Incorrect arguments specified;\
 expected 1 or 2, obtained {0} of value: {1}'.format(str(len(args)), str(args)))




'''
Segmentation encoding format converstion utilities.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from collections import Counter
from .util.lang import enum
# pylint: disable=C0103


BoundaryFormat = enum(position='position', mass='mass', sets='sets')


def convert_positions_to_masses(positions):
    '''
    Convert an ordered sequence of boundary position labels into a
    sequence of segment masses, e.g., ``[1,1,1,1,1,2,2,2,3,3,3,3,3]`` becomes
    ``[5,3,5]``.
    
    :param segments: Ordered sequence of which segments a unit belongs to.
    :type segments: tuple
    
    .. deprecated:: 1.0
    '''
    counts = Counter(positions)
    masses = list()
    for i in range(1, max(counts.keys()) + 1):
        masses.append(counts[i])
    return tuple(masses)


def convert_masses_to_positions(masses):
    '''
    Converts a sequence of segment masses into an ordered sequence of section
    labels for each unit, e.g., ``[5,3,5]`` becomes
    ``[1,1,1,1,1,2,2,2,3,3,3,3,3]``.
    
    :param masses: Segment mass sequence.
    :type masses: tuple
    '''
    sequence = list()
    for i, mass in enumerate(masses):
        sequence.extend([i + 1] * mass)
    return tuple(sequence)


def boundary_string_from_masses(masses):
    '''
    Creates a "boundary string", or sequence of boundary type sets from a list of segment masses, e.g., ``[5,3,5]`` becomes
    ``[(),(),(),(),(1),(),(),(1),(),(),(),()]``.
    

    :param masses: Segmentation masses.
    :type masses: tuple
    '''
    string = [set() for _ in xrange(0, sum(masses) - 1)]
    # Iterate over each position
    pos = 0
    for mass in masses:
        cur_pos = pos + mass - 1
        if cur_pos < len(string):
            string[cur_pos].add(1)
        pos += mass
    # Return
    return tuple([frozenset(pb) for pb in string])




'''
Inter-coder agreement statistic Fleiss' Pi.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from decimal import Decimal
from . import __fnc_metric__, __actual_agreement_linear__


def __fleiss_pi_linear__(items_masses, **kwargs):
    # pylint: disable=C0103,R0914,W0142
    metric_kwargs = dict(kwargs)
    metric_kwargs['return_parts'] = True
    # Arguments
    return_parts = kwargs['return_parts']
    # Check that there are an equal number of items for each coder
    num_items = len(items_masses.values()[0].keys())
    if len([True for coder_segs in items_masses.values() \
            if len(coder_segs.values()) is not num_items]) > 0:
        raise Exception('Unequal number of items contained.')
    # Initialize totals
    all_numerators, all_denominators, _, coders_boundaries = \
        __actual_agreement_linear__(items_masses, **metric_kwargs)
    # Calculate Aa
    A_a = Decimal(sum(all_numerators)) / sum(all_denominators)
    # Calculate Ae
    p_e_segs = list()
    for boundaries_info in coders_boundaries.values():
        for item in boundaries_info:
            boundaries, total_boundaries = item
            p_e_seg = Decimal(boundaries) / total_boundaries
            p_e_segs.append(p_e_seg)
    # Calculate P_e_seg
    P_e_seg = Decimal(sum(p_e_segs)) / len(p_e_segs)
    A_e = (P_e_seg ** 2)
    # Calculate pi
    pi = (A_a - A_e) / (Decimal('1') - A_e)
    # Return
    if return_parts:
        return A_a, A_e
    else:
        return pi


def fleiss_pi_linear(dataset, **kwargs):
    '''
    Calculates Fleiss' :math:`\pi` (or multi-:math:`\pi`), originally proposed in
    [Fleiss1971]_, and is equivalent to Siegel and Castellan's :math:`K`
    [SiegelCastellan1988]_.  For 2 coders, this is equivalent to Scott's :math:`\pi`
    [Scott1955]_.
    '''
    # pylint: disable=W0142
    return __fnc_metric__(__fleiss_pi_linear__, dataset, **kwargs)





'''
Data I/O package.  Used to import and export data to and from TSV and JSON
files.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
# pylint: disable=C0103
import os
import csv
import json
import copy
from collections import defaultdict
from .tsv import input_linear_mass_tsv
from .jsonutils import input_linear_mass_json, Field
from ..format import BoundaryFormat

FILETYPE_TSV  = 'tsv'
FILETYPE_JSON = 'json'

EXT = 'ext'
FNC = 'fnc'
FILETYPES           = {FILETYPE_TSV  : {EXT : ['.tsv', '.csv'],
                                        FNC : input_linear_mass_tsv},
                       FILETYPE_JSON : {EXT : ['.json', '.jsn'],
                                        FNC : input_linear_mass_json}}
FILETYPES_DEFAULT   = FILETYPE_JSON


class Dataset(defaultdict):
    '''
    Represents a set of texts (i.e., items) that have been segmentations by coders.
    '''
    # pylint: disable=R0903
    
    def __init__(self, item_coder_data=None, properties=None,
                 boundary_types=None, boundary_format=BoundaryFormat.mass):
        '''
        Initialize.
        '''
        defaultdict.__init__(self, dict)
        self.properties = dict()
        self.boundary_format = boundary_format
        # Masses
        if item_coder_data is not None and item_coder_data is not dict:
            defaultdict.update(self, item_coder_data)
        # Properties
        if properties is not None:
            self.properties.update(properties)
        # Boundary types
        if boundary_types is not None:
            self.boundary_types = set(boundary_types)
        else:
            self.boundary_types = set([1])
        # Coders
        self.coders = set()
        # Populate coders
        for coder_masses in defaultdict.values(self):
            for coder in coder_masses.keys():
                self.coders.add(coder)
    
    def __iadd__(self, other, prepend_item=None):
        '''
        Add one dataset's data to this dataset
        '''
        # Combine item codings
        for item, codings in other.items():
            if prepend_item is not None:
                item_parts = list()
                item_parts.extend(prepend_item)
                item_parts.append(item)
                item = ','.join(item_parts)
            if item not in self:
                self[item] = dict()
            # For each coder, add ttheir coding
            for coder, item_masses in codings.items():
                self.coders.add(coder)
                if coder not in self[item]:
                    self[item][coder] = item_masses
                else:
                    raise DataIOError('Duplicate coders of same name \
%(coder)s found for item %(item)s' % {'coder' : coder, 'item' : item})
        return self
    
    def __add__(self, other):
        dataset = copy.deepcopy(self)
        dataset += other
        return dataset


def name_from_filepath(filepath):
    '''
    Creates a default coder name from a filename.
    '''
    name = os.path.split(filepath)[1]
    name_basic = os.path.splitext(name)[0]
    name = name_basic if len(name_basic) > 0 else name
    return name


class DataIOError(Exception):
    '''
    Indicates that an input processing error has occurred.
    '''
    
    def __init__(self, message, exception=None):
        '''
        Initializer.
        
        :param message: Explanation for the exception.
        :type message: str
        '''
        Exception.__init__(self, message, exception)


def load_nested_folders_dict(containing_dir, filetype, dataset=None,
                             prepend_item=list()):
    '''
    Loads TSV files from a file directory structure, which reflects the
    directory structure in nested :func:`dict` with each directory name
    representing a key in these :func:`dict`.
    
    :param containing_dir: Root directory containing sub-directories which 
                           contain segmentation files.
    :param filetype:       File type to load (e.g., json or tsv).
    :type containing_dir: str
    :type filetype: str

    '''
    # pylint: disable=R0914
    # Create empty dataset
    if dataset is None:
        dataset = Dataset()
    # Vars
    allowable_extensions = list(FILETYPES[filetype][EXT])
    fnc_load = FILETYPES[filetype][FNC]
    datafile_found = False
    # List of entries
    files = dict()
    dirs = dict()
    # For each filesystem item
    for name in os.listdir(containing_dir):
        path = os.path.join(containing_dir, name)
        # Found a directory
        if os.path.isdir(path):
            dirs[name] = path
        # Found a file
        elif os.path.isfile(path):
            name, ext = os.path.splitext(name)
            if len(ext) > 0 and ext.lower() in allowable_extensions:
                files[name] = path
                datafile_found = True
    # If a data file was found
    if datafile_found:
        # If TSV files were found, load
        for name, filepath in files.items():
            other = fnc_load(filepath)
            dataset.__iadd__(other, prepend_item=prepend_item)
    # If only dirs were found, recurse
    for name, dirpath in dirs.items():
        new_prepend_item = list(prepend_item)
        new_prepend_item.append(name)
        # Recurse
        load_nested_folders_dict(dirpath,
                                 filetype,
                                 dataset=dataset,
                                 prepend_item=new_prepend_item)
    return dataset




'''
Implementation of the WindowDiff segmentation evaluation metric described in
[PevznerHearst2002]_ with an optional modification to fix incorrect error
counting at the beginning and end of segmentations provided by 
[LamprierEtAl2007]_.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from decimal import Decimal
from . import __compute_window_size__, WINDOW_METRIC_DEFAULTS
from ..format import (BoundaryFormat, convert_masses_to_positions,
                      convert_positions_to_masses)
from ..util import __fnc_metric__, SegmentationMetricError


WINDOWDIFF_METRIC_DEFAULTS = dict(WINDOW_METRIC_DEFAULTS)
WINDOWDIFF_METRIC_DEFAULTS.update({
    'lamprier_et_al_2007_fix' : False
    })


def __create_paired_window__(hypothesis, reference, window_size,
                             lamprier_et_al_2007_fix):
    '''
    Create a set of pairs of units from each segmentation to go over using a
    window.
    '''
    phantom_size = 0
    if lamprier_et_al_2007_fix is False:
        units_ref_hyp = zip(reference, hypothesis)
    else:
        phantom_size = window_size
        phantom_size = 1 if phantom_size <= 0 else phantom_size
        phantom = tuple([0] * phantom_size)
        units_ref_hyp = zip(phantom + reference + phantom,
                            phantom + hypothesis + phantom)
    return units_ref_hyp, phantom_size


def __window_diff__(hypothesis, reference, window_size, one_minus,
                    boundary_format, return_parts, fnc_round,
                    lamprier_et_al_2007_fix):
    '''
    Calculates the WindowDiff segmentation evaluation metric score for a
    hypothetical segmentation against a reference segmentation for a given
    window size.  The standard method of calculating the window size
    is performed a window size is not specified.
    
    :param hypothesis:     Hypothesis segmentation section labels
                                        sequence.
    :param reference:      Reference segmentation section labels
                                        sequence.
    :param window_size:              The size of the window that is slid over \
                                        the two segmentations used to count \
                                        mismatches (default is None and will \
                                        use the average window size)
    :param one_minus:                Return 1-WindowDiff to make it no longer \
                                         a penalty-metric.
    :param lamprier_et_al_2007_fix:  Apply a fix for improperly counted errors \
                                        at the beginning and end of \
                                        segmentations, provided by \
                                        _[LamprierEtAl2007].
    :param convert_from_masses:      Convert the segmentations provided from \
                                        masses into positions.
    :type hypothesis: list
    :type reference: list
    :type window_size: int
    :type one_minus: bool
    :type lamprier_et_al_2007_fix: bool
    :type convert_from_masses: bool
    
    .. note:: See :func:`segeval.convert_masses_to_positions` for an example of
              the input format.
    '''
    # pylint: disable=C0103,R0913,R0914,R0912
    # Convert from masses into positions 
    if boundary_format == BoundaryFormat.mass:
        reference  = convert_masses_to_positions(reference)
        hypothesis = convert_masses_to_positions(hypothesis)
    elif boundary_format != BoundaryFormat.position:
        raise SegmentationMetricError('Unsupported boundary format')
    # Check for input errors
    if len(reference) is not len(hypothesis):
        raise SegmentationMetricError(
                    'Reference and hypothesis segmentations differ in position \
length (%(ref)i is not %(hyp)i).' % {'ref' : len(reference),
                                     'hyp' : len(hypothesis)})
    # Compute window size to use if unspecified
    if window_size is None:
        window_size = __compute_window_size__(reference, fnc_round,
                                                BoundaryFormat.position)
    # Create a set of pairs of units from each segmentation to go over using a
    # window
    units_ref_hyp = __create_paired_window__(hypothesis, reference,
                                             window_size,
                                             lamprier_et_al_2007_fix)[0]
    # Slide window over and sum the number of varying windows
    sum_differences = 0
    measurements = len(units_ref_hyp) - window_size
    for i in xrange(0, measurements):
        window = units_ref_hyp[i: i + window_size + 1]
        ref_boundaries = 0
        hyp_boundaries = 0
        # Check that the number of loops is correct
        assert len(window) is window_size + 1
        # For pair in window
        for j in xrange(0, len(window) - 1):
            ref_part, hyp_part = zip(*window[j:j+2])
            # Boundary exists in the reference segmentation
            if ref_part[0] is not ref_part[1]:
                ref_boundaries += 1
            # Boundary exists in the hypothesis segmentation
            if hyp_part[0] is not hyp_part[1]:
                hyp_boundaries += 1
        # If the number of boundaries per segmentation in the window differs
        if ref_boundaries is not hyp_boundaries:
            sum_differences += 1
    # Perform final division
    n = sum(convert_positions_to_masses(reference))
    denominator = n - window_size
    if lamprier_et_al_2007_fix:
        denominator = measurements + 1
    win_diff = Decimal(sum_differences) / denominator
    # Check normalization
    assert denominator is measurements or lamprier_et_al_2007_fix
    # Check value
    assert win_diff <= 1
    if not one_minus:
        if return_parts:
            return sum_differences, denominator
        else:
            return win_diff
    else:
        return Decimal('1.0') - win_diff


def window_diff(*args, **kwargs):
    # pylint: disable=W0142
    return __fnc_metric__(__window_diff__, args, kwargs,
                          WINDOWDIFF_METRIC_DEFAULTS)




'''
Tests the data merge functions and package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
import copy
from .samples import LARGE_DISAGREEMENT, COMPLETE_AGREEMENT


class TestMerge(unittest.TestCase):
    '''
    Test data merge functions.
    '''
    #pylint: disable=R0904,C0103
    
    def test_add(self):
        '''
        Test ``Dataset.add()``.
        '''
        # Output complete and larege disagreement, then merge
        large_disagreement = copy.deepcopy(LARGE_DISAGREEMENT)
        large_disagreement += COMPLETE_AGREEMENT
        self.assertTrue(6, len(large_disagreement.coders))
        self.assertTrue(4, len(large_disagreement))
        self.assertTrue(2, len(large_disagreement['item1']))



'''
Tests segmentation format utility functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from format import (convert_positions_to_masses, convert_masses_to_positions,
                    boundary_string_from_masses)


class TestSegeval(unittest.TestCase):
    '''
    segeval utility function tests.
    '''
    #pylint: disable=R0904,C0103
    
    def test_convert_positions_to_masses(self):
        '''
        Test segment position sequence conversion to masses.
        '''
        #pylint: disable=C0324
        self.assertEqual((5,3,5),
                    convert_positions_to_masses([1,1,1,1,1,2,2,2,3,3,3,3,3]))
    
    
    def test_convert_positions_to_masses_none(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        #pylint: disable=C0324
        self.assertEqual((11,),
                         convert_positions_to_masses([1,1,1,1,1,1,1,1,1,1,1]))
    
    def test_convert_positions_to_masses_all(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        #pylint: disable=C0324
        self.assertEqual((1,1,1,1,1,1,1,1,1,1,1),
                         convert_positions_to_masses([1,2,3,4,5,6,7,8,9,10,11]))
        
      
    def test_convert_masses_to_positions(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        #pylint: disable=C0324
        self.assertEqual((1,1,1,1,1,2,2,2,3,3,3,3,3),
                         convert_masses_to_positions([5,3,5]))
    
    
    def test_convert_masses_to_positions_none(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        #pylint: disable=C0324
        self.assertEqual(convert_masses_to_positions([11]),
                         (1,1,1,1,1,1,1,1,1,1,1))
    
    def test_convert_masses_to_positions_all(self):
        '''
        Test segment position sequence conversion from masses.
        '''
        #pylint: disable=C0324
        self.assertEqual(convert_masses_to_positions([1,1,1,1,1,1,1,1,1,1,1]),
                         (1,2,3,4,5,6,7,8,9,10,11))

    def test_boundary_string_from_masses_none(self):
        '''
        No boundaries.
        '''
        string = boundary_string_from_masses([3])
        self.assertEqual(string, (set(), set()))


    def test_boundary_string_from_masses_full(self):
        '''
        Few boundaries.
        '''
        string = boundary_string_from_masses([1,1,1,1])
        self.assertEqual(string, (set([1]), set([1]), set([1])))


    def test_boundary_string_from_masses_one(self):
        '''
        Few boundaries.
        '''
        string = boundary_string_from_masses([2,3])
        self.assertEqual(string, (set(), set([1]), set(), set()))



        


'''
Tests the data merge functions and package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
import os
import re
from .tsv import input_linear_mass_tsv, input_linear_positions_tsv
from .samples import HEARST_1997_STARGAZER

class TestTsv(unittest.TestCase):
    '''
    Test data merge functions.
    '''
    #pylint: disable=R0904,C0103
    
    test_data_dir = os.path.split(__file__)[0]
    
    def test_input_linear_mass_tsv(self):
        '''
        Test mass TSV file input.
        '''
        tsv_file = os.path.join(self.test_data_dir, 'hearst1997.tsv')
        dataset = input_linear_mass_tsv(tsv_file)
        self.assertEqual(dataset['hearst1997'],
                         HEARST_1997_STARGAZER['stargazer'])
      
    def test_input_linear_positions_tsv(self):
        '''
        Test position TSV file input.
        '''
        tsv_file = os.path.join(self.test_data_dir, 'hearst1997_positions.csv')
        dataset = input_linear_positions_tsv(tsv_file, delimiter=',')
        self.assertEqual(dataset['hearst1997_positions'],
                         HEARST_1997_STARGAZER['stargazer'])



'''
Tests the data merge functions and package.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
import os
import re
from . import DataIOError
from .jsonutils import (output_linear_mass_json, input_linear_mass_json,
    __write_json__, Field)
from .samples import HEARST_1997_STARGAZER


class TestJsonUtils(unittest.TestCase):
    '''
    Test data merge functions.
    '''
    #pylint: disable=R0904,C0103
    test_data_dir = os.path.split(__file__)[0]
    
    def test_output_linear_mass_json(self):
        '''
        Test ``Dataset.add()``.
        '''
        # Output specific file
        file_path_new = os.path.join(self.test_data_dir, 'hearst1997_test.json')
        file_path_existing = os.path.join(self.test_data_dir, 'hearst1997.json')
        output_linear_mass_json(file_path_new, HEARST_1997_STARGAZER)
        self.assertEqual(re.sub(r'\s+', '', open(file_path_new).read()),
                         re.sub(r'\s', '',open(file_path_existing).read()))
        os.remove(file_path_new)
        self.assertFalse(os.path.exists(file_path_new))
        # Output to folder
        file_path_new = os.path.join(self.test_data_dir, 'output.json')
        output_linear_mass_json(self.test_data_dir, HEARST_1997_STARGAZER)
        self.assertEqual(re.sub(r'\s+', '', open(file_path_new).read()),
                         re.sub(r'\s', '',open(file_path_existing).read()))
        os.remove(file_path_new)
        self.assertFalse(os.path.exists(file_path_new))
        
    def test_input_linear_mass_json(self):
        '''
        Test mass JSON file input.
        '''
        json_file = os.path.join(self.test_data_dir, 'hearst1997.json')
        dataset = input_linear_mass_json(json_file)
        self.assertEqual(dataset, HEARST_1997_STARGAZER)

    def test_input_exception(self):
        file_path = os.path.join(self.test_data_dir, 'hearst1997.tsv')
        self.assertRaises(DataIOError, input_linear_mass_json, file_path)

    def test_input_type_exception(self):
        file_path_new = os.path.join(self.test_data_dir, 'hearst1997_test.json')
        __write_json__(file_path_new,
            {Field.segmentation_type : 'incorrect',
             Field.items : {}})
        self.assertRaises(DataIOError, input_linear_mass_json, file_path_new)
        os.remove(file_path_new)


        







'''
Python-language utils.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''

def enum(*sequential, **named):
    '''
    http://stackoverflow.com/a/1695250/2134
    '''
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)



'''
Tests the WindowDiff evaluation metric.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from decimal import Decimal
from .windowdiff import window_diff
from ..data.samples import (KAZANTSEVA2012_G5, KAZANTSEVA2012_G2, 
    COMPLETE_AGREEMENT, LARGE_DISAGREEMENT)
from ..compute import summarize
from ..format import BoundaryFormat
from ..util import SegmentationMetricError
from ..util.test import TestCase
from ..data.samples import HEARST_1997_STARGAZER, HYPOTHESIS_STARGAZER


class TestWindowDiffPositions(TestCase):
    '''
    Test WindowDiff.
    '''
    # pylint: disable=R0904,C0324
    kwargs = {'boundary_format' : BoundaryFormat.position,
              'lamprier_et_al_2007_fix' : False}

    def test_one_minus_var(self):
        '''
        Test one minus.
        '''
        value = window_diff([2, 3, 6], [2, 2, 7], one_minus=True)
        self.assertAlmostEqual(Decimal('0.77777777'), value)
        
    def test_identical(self):
        '''
        Test whether identical segmentations produce 0.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        self.assertEqual(window_diff(a, b, **self.kwargs), 0.0)

    def test_no_boundaries(self):
        '''
        Test whether no segments versus some segments produce 1.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertAlmostEqual(window_diff(a, b, **self.kwargs),
                         Decimal('0.3636363636363636363636363636'))
        self.assertAlmostEqual(window_diff(b, a, **self.kwargs),
                         Decimal('1'))

    def test_all_boundaries(self):
        '''
        Test whether all segments versus some segments produces 0.833
        erroneous windows.
        '''
        # pylint: disable=C0324,C0103
        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertAlmostEqual(window_diff(a, b, **self.kwargs),
                         Decimal('1.0'))
        self.assertAlmostEqual(window_diff(b, a, **self.kwargs),
                         Decimal('1.0'))

    def test_all_and_no_boundaries(self):
        '''
        Test whether all segments versus no segments produces 1.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        self.assertAlmostEqual(window_diff(a, b, **self.kwargs), Decimal('1'))
        self.assertAlmostEqual(window_diff(b, a, **self.kwargs), Decimal('1'))

    def test_translated_boundary(self):
        '''
        Test mis-alignment.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertAlmostEqual(window_diff(a, b, **self.kwargs), # k = 2
                               Decimal(2.0/11.0))
        self.assertAlmostEqual(window_diff(b, a, **self.kwargs),
                               Decimal(2.0/11.0))
    
    def test_extra_boundary(self):
        '''
        Test extra boundary.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertAlmostEqual(window_diff(a, b, **self.kwargs),
                               Decimal(2.0/11.0))
        self.assertAlmostEqual(window_diff(b, a, **self.kwargs),
                               Decimal(2.0/11.0))
    
    def test_full_miss_and_misaligned(self):
        '''
        Test whether a full miss and a translated boundary out of 4 produces
        0.25. 
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertAlmostEqual(window_diff(a, b, **self.kwargs),
                         Decimal('0.2727272727272727272727272727'))
        self.assertAlmostEqual(window_diff(b, a, **self.kwargs),
                         Decimal('0.2727272727272727272727272727'))
    
    def test_fn_vs_fp(self):
        '''
        Test the difference between FP and FN.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertAlmostEqual(window_diff(a, b, **self.kwargs),
                         Decimal('0.1818181818181818181818181818'))
        self.assertAlmostEqual(window_diff(b, a, **self.kwargs),
                         Decimal('0.1818181818181818181818181818'))
        a = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        b = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        self.assertAlmostEqual(window_diff(a, b, **self.kwargs),
                         Decimal('0.1818181818181818181818181818'))
        self.assertAlmostEqual(window_diff(b, a, **self.kwargs),
                         Decimal('0.1818181818181818181818181818'))
    
    def test_parts(self):
        '''
        Test parts.
        '''
        # pylint: disable=C0324,C0103
        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        metric_kwargs = dict(self.kwargs)
        metric_kwargs['return_parts'] = True
        self.assertEqual(window_diff(a, b, **metric_kwargs),
                         (11, 11))
    
    def test_format_exception(self):
        '''
        Test format exception.
        '''
        a = [2, 3, 6]
        b = [2, 2, 7]
        self.assertRaises(SegmentationMetricError, window_diff,
                          a, b, boundary_format=BoundaryFormat.sets)

    def test_mass_exception(self):
        '''
        Test length mismatch exception.
        '''
        a = [2, 2, 7]
        b = [2, 2, 8]
        self.assertRaises(SegmentationMetricError, window_diff, a, b)

    def test_one_minus(self):
        '''
        Test whether no segments versus some segments produce 1.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        metric_kwargs = dict(self.kwargs)
        metric_kwargs['one_minus'] = True
        self.assertAlmostEqual(window_diff(a, b, **metric_kwargs),
                         Decimal('0.6363636363636363636363636364'))
        self.assertAlmostEqual(window_diff(b, a, **metric_kwargs),
                         Decimal('0'))


class TestWindowDiffMasses(TestCase):
    '''
    Test WindowDiff.
    '''
    # pylint: disable=R0904,C0324
    kwargs = {'boundary_format' : BoundaryFormat.mass,
              'lamprier_et_al_2007_fix' : False}

    def test_thesis_all_and_none(self):
        '''
        Test paper example A vs B
        '''
        reference = [11]
        hypothesis = [1,1,1,1,1,1,1,1,1,1,1]
        # Test normal
        actual = window_diff(hypothesis, reference, **self.kwargs)
        self.assertAlmostEqual(1.0, float(actual))
    
    def test_scaiano_paper_b(self):
        '''
        Test paper example A vs B
        '''
        reference = [6,6]
        hypothesis = [12]
        # Test normal
        actual = window_diff(hypothesis, reference, **self.kwargs)
        self.assertAlmostEqual(3.0/9.0, float(actual))
    
    def test_scaiano_paper_c(self):
        '''
        Test paper example A vs C
        '''
        reference = [6,6]
        hypothesis = [5,7]
        # Test normal
        actual = window_diff(hypothesis, reference, **self.kwargs)
        self.assertAlmostEquals(2.0/9.0, float(actual))
    
    def test_scaiano_paper_d(self):
        '''
        Test paper example A vs D
        '''
        reference = [6,6]
        hypothesis = [1,5,6]
        # Test normal
        actual = window_diff(hypothesis, reference, **self.kwargs)
        self.assertAlmostEquals(1.0/9.0, float(actual))
    
    def test_scaiano_paper_e(self):
        '''
        Test paper example A vs E
        '''
        reference = [6,6]
        hypothesis = [5,1,1,5]
        # Test normal
        actual = window_diff(hypothesis, reference, **self.kwargs)
        self.assertAlmostEquals(5.0/9.0, float(actual))


class TestPairwiseWindowDiff(TestCase):
    # pylint: disable=R0904,E1101,W0232
    '''
    Test pairwise WindowDiff.
    '''
    
    def test_kazantseva2012_g5(self):
        '''
        Calculate permuted pairwise WindowDiff on Group 5 from the dataset
        collected in [KazantsevaSzpakowicz2012]_.
        '''
        self.assertAlmostEquals(summarize(window_diff(KAZANTSEVA2012_G5,
                                              lamprier_et_al_2007_fix=False)),
                         (Decimal('0.42514977232721135'),
                          Decimal('0.14960495739111837'),
                          Decimal('0.02238164327599834'),
                          Decimal('0.02159361560546617'),
                          48))
        self.assertAlmostEquals(summarize(window_diff(KAZANTSEVA2012_G5,
                                              lamprier_et_al_2007_fix=True)),
                         (Decimal('0.3981624488837195'),
                          Decimal('0.1499358110615409'),
                          Decimal('0.022480747438682093'),
                          Decimal('0.021641370219386377'),
                          48))
    def test_pair_g5(self):
        '''
        Test a comparison that is troublesome when using lamprier_et_al_2007_fix 
        '''
        # pylint: disable=C0103
        wd = window_diff(KAZANTSEVA2012_G5['ch4']['an2'],
                         KAZANTSEVA2012_G5['ch4']['an1'],
                         lamprier_et_al_2007_fix=True)
        self.assertAlmostEqual(0.76363636, float(wd))
    
    def test_kazantseva2012_g2(self):
        '''
        Calculate mean permuted pairwise WindowDiff on Group 2 from the dataset
        collected in [KazantsevaSzpakowicz2012]_.
        '''
        self.assertAlmostEquals(summarize(window_diff(KAZANTSEVA2012_G2,
                                              lamprier_et_al_2007_fix=False)),
                         (Decimal('0.3257163091933661553914718469'),
                          Decimal('0.1586420969856167116081811670'),
                          Decimal('0.02516731493599381893573908435'),
                          Decimal('0.01448197584815743151147537110'),
                          120))
        self.assertAlmostEquals(summarize(window_diff(KAZANTSEVA2012_G2,
                                              lamprier_et_al_2007_fix=True)),
                         (Decimal('0.2745037663246318112728760428'),
                          Decimal('0.1093940158628282748721971010'),
                          Decimal('0.01196705070659672423205913742'),
                          Decimal('0.009986261690691502298962284136'),
                          120))
    
    def test_large_disagreement(self):
        '''
        Calculate mean permuted pairwise WindowDiff on a theoretical dataset
        containing large disagreement.
        '''
        val = summarize(window_diff(LARGE_DISAGREEMENT,
                                    lamprier_et_al_2007_fix=False))
        self.assertAlmostEquals(val,
                        (Decimal('1'),
                         Decimal('0'),
                         Decimal('0'),
                         Decimal('0'),
                         8))
    
    def test_complete_agreement(self):
        '''
        Calculate mean permuted pairwise WindowDiff on a theoretical dataset
        containing complete agreement.
        '''
        self.assertAlmostEquals(summarize(window_diff(COMPLETE_AGREEMENT,
                                                lamprier_et_al_2007_fix=False)),
                         (0.0,
                          0.0,
                          0.0,
                          0.0,
                          48))
        self.assertAlmostEquals(summarize(window_diff(COMPLETE_AGREEMENT,
                                                lamprier_et_al_2007_fix=True)),
                          (0.0,
                           0.0,
                           0.0,
                           0.0,
                           48))

    def test_wd_datasets(self):
        '''
        Test WD upon two datasets.
        '''
        hypothesis = HYPOTHESIS_STARGAZER
        reference = HEARST_1997_STARGAZER
        value = window_diff(hypothesis, reference)

        # Precision
        self.assertAlmostEquals(float(value['stargazer,h1,1']), 0.31578947)
        self.assertAlmostEquals(float(value['stargazer,h2,1']), 0.47368421)
        self.assertAlmostEquals(float(value['stargazer,h1,2']), 0.42105263)
        self.assertAlmostEquals(float(value['stargazer,h2,2']), 0.47368421)




'''
Weighting functions for edit operations produced by boundary edit distance.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from decimal import Decimal


def weight_a(additions):
    '''
    Default unweighted weighting function for addition edit operations.
    '''
    return len(additions)


def weight_s(substitutions, max_s, min_s=1):
    '''
    Unweighted weighting function for substitution edit operations.
    '''
    # pylint: disable=W0613,C0103
    return len(substitutions)


def weight_s_scale(substitutions, max_s, min_s=1):
    '''
    Default weighting function for substitution edit operations by the distance between ordinal boundary types.
    '''
    # pylint: disable=W0613,C0103
    return weight_t_scale(substitutions, max_s - min_s + 1)


def weight_t(transpositions, max_n):
    '''
    Unweighted weighting function for transposition edit operations.
    '''
    # pylint: disable=W0613,C0103
    return len(transpositions)


def weight_t_scale(transpositions, max_n):
    '''
    Default weighting function for transposition edit operations by the distance that transpositions span.
    '''
    numerator = 0
    if isinstance(transpositions, list):
        for transposition in transpositions:
            numerator += abs(transposition[0] - transposition[1])
        return Decimal(numerator) / max_n




'''
Tests overall package functions and classes.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
import segeval
from segeval import (KAZANTSEVA2012_G5, actual_agreement_linear,
    artstein_poesio_bias_linear, fleiss_kappa_linear, fleiss_pi_linear,
    precision, recall, pk, window_diff, summarize, fmeasure,
    boundary_confusion_matrix, boundary_statistics, boundary_edit_distance,
    compute_window_size, boundary_string_from_masses,
    convert_positions_to_masses, convert_masses_to_positions)


class TestModule(unittest.TestCase):
    '''
    Test module loading.
    '''
    #pylint: disable=R0904,C0103
    def test_dir(self):
        self.assertEquals(set(dir(segeval)),
            set(['Average', 'BoundaryFormat', 'COMPLETE_AGREEMENT',
             'ConfusionMatrix', 'Dataset', 'Field', 'KAZANTSEVA2012_G2',
             'KAZANTSEVA2012_G5', 'LARGE_DISAGREEMENT', 'HYPOTHESIS_STARGAZER',
             'HEARST_1997_STARGAZER', '__all__', '__doc__',
             '__docformat__', '__file__', '__name__', '__package__',
             '__path__', '__path__', '__project__', '__version__', 'actual_agreement_linear',
             'artstein_poesio_bias_linear', 'boundary_confusion_matrix',
             'boundary_edit_distance', 'boundary_statistics',
             'boundary_similarity', 'segmentation_similarity',
             'boundary_string_from_masses', 'compute_window_size',
             'convert_masses_to_positions', 'convert_positions_to_masses',
             'fleiss_kappa_linear', 'fleiss_pi_linear', 'fmeasure',
             'input_linear_mass_json', 'input_linear_mass_tsv',
             'load_nested_folders_dict', 'output_linear_mass_json', 'pk',
             'precision', 'recall', 'summarize', 'weight_t', 'weight_s_scale',
             'weight_t_scale', 'weight_s', 'weight_a', 'window_diff']))

    def test_get_attr(self):
        self.assertEquals(segeval.__getattr__('__package__'), 'segeval')


class TestExamples(unittest.TestCase):
    '''
    Example of segeval function usage.
    '''
    #pylint: disable=R0904,C0103

    masses_an1 = KAZANTSEVA2012_G5['ch1']['an1']
    masses_an2 = KAZANTSEVA2012_G5['ch1']['an2']

    
    def test_actual_agreement_linear(self):
        '''
        Test actual_agreement_linear.
        '''
        #pylint: disable=C0324

        self.assertAlmostEquals(Decimal('0.25645756'),
            actual_agreement_linear(KAZANTSEVA2012_G5))

    
    def test_artstein_poesio_bias_linear(self):
        '''
        Test artstein_poesio_bias_linear.
        '''
        #pylint: disable=C0324

        self.assertAlmostEquals(Decimal('0.00841453'),
            artstein_poesio_bias_linear(KAZANTSEVA2012_G5))

    
    def test_fleiss_kappa_linear(self):
        '''
        Test fleiss_kappa_linear.
        '''
        #pylint: disable=C0324

        self.assertAlmostEquals(Decimal('0.23740302'),
            fleiss_kappa_linear(KAZANTSEVA2012_G5))

    
    def test_fleiss_pi_linear(self):
        '''
        Test fleiss_pi_linear.
        '''
        #pylint: disable=C0324

        self.assertAlmostEquals(Decimal('0.23076438'),
            fleiss_pi_linear(KAZANTSEVA2012_G5))

    
    def test_precision(self):
        '''
        Test precision.
        '''
        #pylint: disable=C0324
        cm = boundary_confusion_matrix(self.masses_an1, self.masses_an2)
        self.assertAlmostEquals(Decimal('0.14285714'), precision(cm))

    
    def test_recall(self):
        '''
        Test recall.
        '''
        #pylint: disable=C0324
        cm = boundary_confusion_matrix(self.masses_an1, self.masses_an2)
        self.assertAlmostEquals(Decimal('0.14285714'), recall(cm))

    
    def test_fmeasure(self):
        '''
        Test fmeasure.
        '''
        #pylint: disable=C0324
        cm = boundary_confusion_matrix(self.masses_an1, self.masses_an2)
        self.assertAlmostEquals(Decimal('0.25000000'), fmeasure(cm))

    
    def test_boundary_statistics(self):
        '''
        Test boundary_statistics.
        '''
        #pylint: disable=C0324
        self.assertEquals(24, len(boundary_statistics(KAZANTSEVA2012_G5)))


    def test_boundary_edit_distance(self):
        '''
        Test boundary_edit_distance.
        '''
        #pylint: disable=C0324
        edits = ([(1, 'b'), (1, 'b'), (1, 'b')], [], [(9, 10, 1)])
        self.assertEquals(edits, boundary_edit_distance(
            boundary_string_from_masses(self.masses_an1),
            boundary_string_from_masses(self.masses_an2)))

    
    def test_compute_window_size(self):
        '''
        Test compute_window_size.
        '''
        #pylint: disable=C0324
        self.assertEquals(3, compute_window_size(self.masses_an1))
        self.assertEquals(3, compute_window_size(KAZANTSEVA2012_G5))

    
    def test_pk(self):
        '''
        Test pk.
        '''
        #pylint: disable=C0324
        mean, std, var, stderr, count = \
            summarize(pk(KAZANTSEVA2012_G5))
        self.assertAlmostEquals(Decimal('0.35530058'), mean)
        self.assertAlmostEquals(Decimal('0.11001760'), std)
        self.assertAlmostEquals(Decimal('0.01210387'), var)
        self.assertAlmostEquals(Decimal('0.01587967'), stderr)
        self.assertEquals(48, count)

    
    def test_window_diff(self):
        '''
        Test window_diff.
        '''
        #pylint: disable=C0324
        mean, std, var, stderr, count = \
            summarize(window_diff(KAZANTSEVA2012_G5))
        self.assertAlmostEquals(Decimal('0.42514977'), mean)
        self.assertAlmostEquals(Decimal('0.14960495'), std)
        self.assertAlmostEquals(Decimal('0.02238164'), var)
        self.assertAlmostEquals(Decimal('0.02159361'), stderr)
        self.assertEquals(48, count)

    
    def test_boundary_string_from_masses(self):
        '''
        Test boundary_string_from_masses.
        '''
        #pylint: disable=C0324

        self.assertEquals((
            frozenset([]), frozenset([]), frozenset([]), frozenset([]),
            frozenset([]), frozenset([]), frozenset([]), frozenset([]),
            frozenset([]), frozenset([]), frozenset([1]), frozenset([])),
            boundary_string_from_masses(self.masses_an1))


    def test_convert_positions_to_masses(self):
        '''
        Test convert_positions_to_masses.
        '''
        #pylint: disable=C0324
        self.assertEquals((4,2), convert_positions_to_masses([1,1,1,1,2,2]))


    def test_convert_masses_to_positions(self):
        '''
        Test convert_masses_to_positions.
        '''
        #pylint: disable=C0324

        self.assertAlmostEquals((1,1,1,1,2,2),
            convert_masses_to_positions((4,2)))



'''
Math utility functions.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from __future__ import division
from decimal import Decimal


def mean(values):
    '''
    Calculates the mean of a list of numeric values.
    
    :param values: List of numeric values.
    :type values: list
    
    :returns: Mean.
    :rtype: :class:`decimal.Decimal`
    '''
    summation = Decimal(0)
    for value in values:
        summation += value
    if len(values) > 0:
        return summation / len(values)
    else:
        return 0
    
    
def var(values):
    '''
    Calculates the population variance of a list of numeric values.
    
    :param values: List of numeric values.
    :type values: list
    
    :returns: Variance.
    :rtype: :class:`decimal.Decimal`
    '''
    mean_value = mean(values)
    summation = Decimal(0)
    for value in values:
        summation += (value - mean_value) ** 2
    return summation / len(values)


def std(values):
    '''
    Calculates the population standard deviation of a list of numeric values.
    
    :param values: List of numeric values.
    :type values: list
    
    :returns: Standard deviation.
    :rtype: :class:`decimal.Decimal`
    '''
    return var(values).sqrt()


def stderr(values):
    '''
    Calculates the population standard error of the mean of a list of numeric
    values.
    
    :param values: List of numeric values.
    :type values: list
    
    :returns: Standard error of the mean.
    :rtype: :class:`decimal.Decimal`
    '''
    return std(values) / Decimal(len(values)).sqrt()


__all__ = []




'''
Tests the segmentation versions of Cohen's and Fleiss' Kappa.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from .kappa import fleiss_kappa_linear
from ..data.samples import (KAZANTSEVA2012_G5, KAZANTSEVA2012_G2, 
    COMPLETE_AGREEMENT, LARGE_DISAGREEMENT)


class TestKappa(unittest.TestCase):
    '''
    Test segmentation versions of Cohen's Kappa and Fleiss' Multi-Kappa.
    '''
    # pylint: disable=R0904

    def test_fleiss_kappa_linear_g5(self):
        '''
        Test Kappa upon Group 5 of Kazantseva (2012) data.
        '''
        data = KAZANTSEVA2012_G5
        self.assertEqual(fleiss_kappa_linear(data),
                         Decimal('0.2374030241245084051985691283'))

    def test_fleiss_kappa_linear_g5_ch1(self):
        '''
        Test Kappa upon Group 5, Chapter 1, of Kazantseva (2012) data.
        '''
        data = KAZANTSEVA2012_G5['ch1']
        data = {'ch1' : data}
        self.assertEqual(fleiss_kappa_linear(data),
                         Decimal('0.1940298507462686567164179104'))

    def test_fleiss_kappa_linear_g2(self):
        '''
        Test Kappa upon Group 2 of Kazantseva (2012) data.
        '''
        data = KAZANTSEVA2012_G2
        self.assertEqual(fleiss_kappa_linear(data),
                         Decimal('0.4068521352341765453201908783'))

    def test_fleiss_kappa_linear_g2_ch2(self):
        '''
        Test Kappa upon Group 2, Chapter 2, of Kazantseva (2012) data.
        '''
        data = KAZANTSEVA2012_G2['ch2']
        data = {'ch2' : data}
        self.assertEqual(fleiss_kappa_linear(data),
                         Decimal('0.5197205281147376181221097781'))
        
    def test_fleiss_kappa_disagree(self):
        '''
        Test Kappa upon a hypothetical dataset containing large disagreement.
        '''
        data = LARGE_DISAGREEMENT
        self.assertEqual(fleiss_kappa_linear(data),
                         Decimal('-0.05952156715012243360331512524'))
    
    def test_fleiss_kappa(self):
        '''
        Test Cohen's and Fleiss' Kappa.
        '''
        data1 = {'i1' : {'c1' : [2, 8, 2, 1],
                         'c2' : [2, 1, 7, 2, 1]}}
        kappa1  = fleiss_kappa_linear(data1)
        kappa1f = fleiss_kappa_linear(data1)
        self.assertEqual(kappa1,
                         Decimal('0.7096774193548387096774193548'))
        self.assertEqual(kappa1, kappa1f)
        data2 = {'i1' : {'c1' : [2, 8, 2, 1],
                         'c2' : [11, 2]}}
        kappa2  = fleiss_kappa_linear(data2)
        kappa2f = fleiss_kappa_linear(data2)
        self.assertEqual(kappa2,
                         Decimal('0.1176470588235294117647058823'))
        self.assertEqual(kappa2, kappa2f)
        self.assertTrue(kappa2 < kappa1)
    
    def test_fleiss_kappa_complete(self):
        '''
        Test Kappa upon a hypothetical dataset containing complete agreement.
        '''
        data_complete = COMPLETE_AGREEMENT
        kappa = fleiss_kappa_linear(data_complete)
        self.assertEqual(kappa, 1.0)
        
    def test_exception_coders(self):
        '''
        Test exception.
        '''
        data = {'i1' : {'c1' : [2, 8, 2, 1]}}
        self.assertRaises(Exception, fleiss_kappa_linear, data)

    def test_exception_items(self):
        '''
        Test exception.
        '''
        data = {'i1' : {'c1' : [2, 8, 2, 1],
                        'c2' : [2, 8, 2, 1]},
                'i2' : {'c1' : [2, 1, 7, 2, 1],
                        'c2' : [2, 8, 2, 1],
                        'c3' : [2, 8, 2, 1]}}
        self.assertRaises(Exception, fleiss_kappa_linear, data)




'''
Tests segmentation similarity (S).

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from .segmentation import segmentation_similarity
from ..util import SegmentationMetricError
from ..data.samples import (HEARST_1997_STARGAZER, HYPOTHESIS_STARGAZER,
    MULTIPLE_BOUNDARY_TYPES, KAZANTSEVA2012_G5)


class TestSegmentation(unittest.TestCase):
    '''
    Test.
    '''
    # pylint: disable=R0904,C0103,C0324
    
    def test_fn(self):
        '''
        Test false negative.
        '''
        value = segmentation_similarity([2, 3, 6], [5, 6])
        self.assertEqual(Decimal('0.9'), value)
    
    def test_fp(self):
        '''
        Test false negative.
        '''
        value = segmentation_similarity([2, 3, 6], [2, 3, 3, 3])
        self.assertEqual(Decimal('0.9'), value)
    
    def test_near_miss(self):
        '''
        Test near miss.
        '''
        value = segmentation_similarity([2, 3, 6], [2, 2, 7])
        self.assertEqual(Decimal('0.95'), value)
    
    def test_one_minus(self):
        '''
        Test one minus.
        '''
        value = segmentation_similarity([2, 3, 6], [2, 2, 7], one_minus=True)
        self.assertEqual(Decimal('0.05'), value)
    
    def test_clustered_fps(self):
        '''
        Test near miss.
        '''
        value = segmentation_similarity([2, 3, 6], [1, 1, 3, 1, 5])
        self.assertEqual(Decimal('0.8'), value)
    
    def test_parts(self):
        '''
        Test false negative.
        '''
        numerator, denominator  = segmentation_similarity([2, 3, 6], [5, 6],
                                                          return_parts=True)
        self.assertEqual(Decimal('9'), numerator)
        self.assertEqual(Decimal('10'), denominator)
    
    def test_format_exception(self):
        '''
        Test false negative.
        '''
        a = [2, 3, 6]
        b = [2, 2, 7]
        self.assertRaises(SegmentationMetricError, segmentation_similarity,
                          a, b, boundary_format=None)

    def test_mass_exception(self):
        '''
        Test false negative.
        '''
        a = [2, 2, 7]
        b = [2, 2, 8]
        self.assertRaises(SegmentationMetricError, segmentation_similarity,
                          a, b)

    def test_s_datasets(self):
        '''
        Test S upon two datasets.
        '''
        hypothesis = HYPOTHESIS_STARGAZER
        reference = HEARST_1997_STARGAZER
        value = segmentation_similarity(hypothesis, reference)

        self.assertAlmostEquals(float(value['stargazer,h1,1']), 0.85)
        self.assertAlmostEquals(float(value['stargazer,h2,1']), 0.725)
        self.assertAlmostEquals(float(value['stargazer,h1,2']), 0.8)
        self.assertAlmostEquals(float(value['stargazer,h2,2']), 0.7)

    def test_s_datasets_return_parts(self):
        '''
        Test S upon two datasets and return fnc parts.
        '''
        hypothesis = HYPOTHESIS_STARGAZER
        reference = HEARST_1997_STARGAZER
        value = segmentation_similarity(hypothesis, reference, return_parts=True)

        self.assertEquals(value['stargazer,h1,1'], (Decimal('17'), 20))

    def test_s_datasets_exception(self):
        '''
        Test S upon two datasets that produces an exception.
        '''
        hypothesis = MULTIPLE_BOUNDARY_TYPES
        reference = HEARST_1997_STARGAZER
        self.assertRaises(SegmentationMetricError, segmentation_similarity, hypothesis, reference)
    
    def test_s_datasets_continue(self):
        '''
        Test S upon two datasets that compares no items.
        '''
        hypothesis = KAZANTSEVA2012_G5
        reference = HEARST_1997_STARGAZER
        value = segmentation_similarity(hypothesis, reference)
        self.assertEqual(value, {})




'''
Data merge tools.  Used to merge multiple files into one.

.. seealso:: File format documentation in: `Segmentation Representation \
Specification <http://nlp.chrisfournier.ca/publications/#seg_spec>`_.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from . import load_files, Dataset, DataIOError
from .jsonutils import output_linear_mass_json


OUTPUT_NAME = 'Merge segmentations operation'
SHORT_NAME  = 'Merge'


def parse(args):
    '''
    Parse this module's metric arguments and perform requested actions.
    '''
    # pylint: disable=C0103
    output = None
        
    # Is a TSV requested?
    if args['output'] is not None:
        # Create a TSV
        output_file = args['output'][0]
        values = load_files(args)
        dataset = Dataset()
        
        for other in values:
            dataset.add(other)
        
        # Output
        output = 'Merged:\n\tFiles:\t%(files)i\n\tCoders:\t%(coders)i\n\t\
Items:\t%(items)i' % {'files'  : len(values),
                      'coders' : len(dataset.coders),
                      'items'  : len(dataset)}
        output_linear_mass_json(output_file,
                                dataset)
    else:
        raise DataIOError('No output file specified.')
    return output


def create_parser(subparsers):
    '''
    Setup a command line parser for this module's metric.
    '''
    parser = subparsers.add_parser('merge',
                                   help=OUTPUT_NAME)
    parser_add_merge_support(parser)
    parser.set_defaults(func=parse)


def parser_add_merge_support(parser):
    '''
    Add support for file input and output parameters to an argument parser.
    
    :param parser: Argument parser
    :type parser: argparse.ArgumentParser
    '''
    from ..data import parser_add_format_support
    
    parser.add_argument('input',
                        type=str,
                        nargs='+',
                        action='store',
                        help='Input files or directories')
    
    parser_add_format_support(parser)




'''
Tests the segmentation version of Arstein and Poesio's bias [ArtsteinPoesio2008]_.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from decimal import Decimal
from .bias import artstein_poesio_bias_linear
from ..data.samples import (KAZANTSEVA2012_G5, KAZANTSEVA2012_G2, 
    COMPLETE_AGREEMENT, LARGE_DISAGREEMENT)


class TestBias(unittest.TestCase):
    '''
    Test Arstein and Poesio's (2008) annotator bias.
    '''
    # pylint: disable=R0904

    def test_bias_g5(self):
        '''
        Test bias upon Group 5 of Kazantseva (2012) data.
        '''
        bias = artstein_poesio_bias_linear(KAZANTSEVA2012_G5)
        self.assertTrue(bias > 0)
        self.assertEqual(bias,
                         Decimal('0.00841453429829254475759269324'))

    def test_bias_g5_ch1(self):
        '''
        Test bias upon Group 5, Chapter 1, of Kazantseva (2012) data.
        '''
        data = KAZANTSEVA2012_G5['ch1']
        data = {'ch1' : data}
        bias = artstein_poesio_bias_linear(data)
        self.assertTrue(bias > 0)
        self.assertEqual(bias,
                         Decimal('0.00390625000000000000000000011'))

    def test_bias_g2(self):
        '''
        Test bias upon Group 2 of Kazantseva (2012) data.
        '''
        bias = artstein_poesio_bias_linear(KAZANTSEVA2012_G2)
        self.assertTrue(bias > 0)
        self.assertEqual(bias,
                         Decimal('0.00821695210923559041105482425'))

    def test_bias_g2_ch2(self):
        '''
        Test bias upon Group 2, Chapter 2, of Kazantseva (2012) data.
        '''
        data = KAZANTSEVA2012_G2['ch2']
        data = {'ch2' : data}
        bias = artstein_poesio_bias_linear(data)
        self.assertTrue(bias > 0)
        self.assertEqual(bias,
                         Decimal('0.00090702947845804988662131528'))

    def test_bias_complete(self):
        '''
        Test bias upon a hypothetical dataset containing complete agreement.
        '''
        bias = artstein_poesio_bias_linear(COMPLETE_AGREEMENT)
        self.assertTrue(bias >= 0)
        self.assertEqual(bias,
                         Decimal('0.01455229356727327645713789012'))
    
    def test_bias_large(self):
        '''
        Test bias upon a hypothetical dataset containing large disagreement.
        '''
        bias = artstein_poesio_bias_linear(LARGE_DISAGREEMENT)
        self.assertTrue(bias >= 0)
        self.assertEqual(bias,
                         Decimal('0.3092215912041560442272265692'))

    def test_parts(self):
        '''
        Test bias upon a hypothetical dataset containing large disagreement.
        '''
        # pylint: disable=C0103
        A_pi_e, A_fleiss_e = artstein_poesio_bias_linear(LARGE_DISAGREEMENT,
                                                return_parts=True)
        self.assertEqual(A_pi_e,
                         Decimal('0.3653993689819338220050043470'))
        self.assertEqual(A_fleiss_e,
                         Decimal('0.05617777777777777777777777776'))
        self.assertEqual(A_pi_e - A_fleiss_e,
                         Decimal('0.3092215912041560442272265692'))

    def test_exception_coders(self):
        '''
        Test exception.
        '''
        data = {'i1' : {'c1' : [2, 8, 2, 1]}}
        self.assertRaises(Exception, artstein_poesio_bias_linear, data)

    def test_exception_items(self):
        '''
        Test exception.
        '''
        data = {'i1' : {'c1' : [2, 8, 2, 1],
                        'c2' : [2, 8, 2, 1]},
                'i2' : {'c1' : [2, 1, 7, 2, 1],
                        'c2' : [2, 8, 2, 1],
                        'c3' : [2, 8, 2, 1]}}
        self.assertRaises(Exception, artstein_poesio_bias_linear, data)





'''
Tests multiple-boundary edit distance.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
import unittest
from .multipleboundary import (boundary_edit_distance, 
    __additions_substitutions__, __additions_substitutions_sets__)


class TestMultipleBoundaries(unittest.TestCase):
    '''
    Test items.
    '''
    # pylint: disable=R0904,C0324,C0103,C0301
    
    def test_additions_substitutions(self):
        '''
        Test the expected functionality of __additions_substitutions__
        '''
        a_i = set([1, 2, 3, 4])
        b_i = set([1, 6])
        
        a = a_i - b_i
        b = b_i - a_i
        d = a_i ^ b_i
        
        self.assertEqual((2, 1),
                         __additions_substitutions__(d, a, b))
    
    def test_additions_substitutions_sets(self):
        '''
        Test the expected functionality of __additions_substitutions_sets__
        '''
        a_i = set([1, 2, 3, 4])
        b_i = set([1, 6])
        
        a = a_i - b_i
        b = b_i - a_i
        d = a_i ^ b_i
        
        self.assertEqual(([(2, 'a'), (3, 'a')], set([(4, 6)])), 
                         __additions_substitutions_sets__(d, a, b))
    
    def test_edit_distance_identity(self):
        '''
        Test a mixed example of additions, substitutions and transpositions.
        '''
        a = [set(), set([1]), set(), set(), set([1]), set(), set(), set(), set(), set()]
        b = a
        
        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([], [], []),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance_compliment(self):
        '''
        Test a mixed example of additions, substitutions and transpositions.
        '''
        a = [set(   ), set(   ), set(   ), set(   ), set(   ), set(   ), set(   ), set(   ), set(   ), set(   )]
        b = [set([1]), set([1]), set([1]), set([1]), set([1]), set([1]), set([1]), set([1]), set([1]), set([1])]
        
        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b'), (1, 'b')], [], []),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance(self):
        '''
        Test a mixed example of additions, substitutions and transpositions.
        '''
        a = [set(), set([1   ]), set(), set(), set([1]), set(   ), set(), set(), set(   ), set()]
        b = [set(), set([2, 3]), set(), set(), set(   ), set([1]), set(), set(), set([3]), set()]
        
        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(3, 'b'), (3, 'b')], [(1, 2)], [(4, 5, 1)]),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance_two_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2]), set(), set(   ), set([2]), set(), set(), set()]
        b = [set(), set(), set(), set(   ), set(), set([2]), set(   ), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(2, 'a')], [], [(5, 6, 2)]),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance_no_two_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set([2]), set(), set(), set([2]), set([2]), set(), set(), set()]
        b = [set(), set(), set(   ), set(), set(), set([2]), set([2]), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(2, 'a')], [], []),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance_two_transpositions_equal(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set(), set([2]), set(   ), set([2]), set(), set(), set()]
        b = [set(), set(), set(), set(), set(   ), set([2]), set(   ), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b)
        self.assertEqual(([(2, 'a')], [], [(4, 5, 2)]),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance_two_substitutions_into_two_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2]), set(), set([1]), set(), set(), set(), set()]
        b = [set(), set(), set(), set([1]), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([], [(2, 1), (1, 2)], []),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance_two_substitutions_into_no_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2]), set(), set([3]), set(), set(), set(), set()]
        b = [set(), set(), set(), set([1]), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([], [(2,1), (3,2)], []),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance_two_substitutions_one_ad_into_one_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2,4]), set(), set([3]), set(), set(), set(), set()]
        b = [set(), set(), set(), set([1]), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([(4, 'a')], [(2, 1), (3, 2)], []),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance_two_substitutions_one_ad_into_no_transpositions(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([2]), set(), set([3]), set(), set(), set(), set()]
        b = [set(), set(), set(), set([1,4]), set(), set([2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([(4, 'b')], [(2,1), (3,2)], []),
                         (additions, substitutions, transpositions))
    
    def test_edit_distance_three_transpositions_overlapping(self):
        '''
        Test two transpositions
        '''
        a = [set(), set(), set(), set([1,2,3]), set(),    set(     ), set(), set(), set(), set()]
        b = [set(), set(), set(), set(       ), set([3]), set([1,2]), set(), set(), set(), set()]
        additions, substitutions, transpositions = boundary_edit_distance(a, b, n_t=3)
        self.assertEqual(([], [], [(3,4,3),(3,5,1),(3,5,2)]),
                         (additions, substitutions, transpositions))




'''
Tests the WindowDiff evaluation metric.

.. moduleauthor:: Chris Fournier <chris.m.fournier@gmail.com>
'''
from decimal import Decimal
from ..compute import summarize
from ..format import BoundaryFormat
from ..window.pk import pk
from ..data.samples import (KAZANTSEVA2012_G5, KAZANTSEVA2012_G2, 
    COMPLETE_AGREEMENT, LARGE_DISAGREEMENT)
from ..util import SegmentationMetricError
from ..util.test import TestCase
from ..data.samples import HEARST_1997_STARGAZER, HYPOTHESIS_STARGAZER


class TestPk(TestCase):
    '''
    Test Pk.
    '''
    # pylint: disable=R0904
    kwargs = {'boundary_format' : BoundaryFormat.position}
    
    def test_one_minus(self):
        '''
        Test one minus.
        '''
        value = pk([2, 3, 6], [2, 2, 7], one_minus=True)
        self.assertAlmostEqual(Decimal('0.77777777'), value)

    def test_return_parts(self):
        '''
        Test one minus.
        '''
        value = pk(KAZANTSEVA2012_G5, return_parts=True)
        self.assertEqual((3, 10), value['ch1,an3,an1'])

    def test_return_parts_dataset(self):
        '''
        Test one minus.
        '''
        value = pk([2, 3, 6], [2, 2, 7], return_parts=True)
        self.assertEqual((2, 9), value)

    def test_identical(self):
        '''
        Test whether identical segmentations produce 0.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        one_minus_kwargs = dict(TestPk.kwargs)
        one_minus_kwargs['one_minus'] = True
        self.assertEqual(pk(a, b, **self.kwargs), 0.0)
        self.assertEqual(pk(a, b, **one_minus_kwargs), 1.0)

    def test_no_boundaries(self):
        '''
        Test whether no segments versus some segments produce 1.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('1.0'))
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.3636363636363636363636363636'))

    def test_all_boundaries(self):
        '''
        Test whether all segments versus some segments produces 7/11 = 0.636
        erroneous windows.
        '''
        # pylint: disable=C0324,C0103
        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.6363636363636363636363636364'))
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('0.6363636363636363636363636364'))

    def test_all_and_no_boundaries(self):
        '''
        Test whether all segments versus no segments produces 1.0.
        '''
        # pylint: disable=C0324,C0103
        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,1,1,1,1,1,1,1,1,1]
        self.assertEqual(pk(a, b, **self.kwargs), 1.0)
        self.assertEqual(pk(b, a, **self.kwargs), 1.0)

    def test_translated_boundary(self):
        '''
        Test whether 2/3 total segments participate in mis-alignment produces
        0.182.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.1818181818181818181818181818'))
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('0.1818181818181818181818181818'))
    
    def test_extra_boundary(self):
        '''
        Test whether 1/3 segments that are non-existent produces 0.091.
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,1,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.09090909090909090909090909091'))
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('0.09090909090909090909090909091'))
    
    def test_full_miss_and_misaligned(self):
        '''
        Test whether a full miss and a translated boundary out of 4 produces
        0.273. 
        '''
        # pylint: disable=C0324,C0103
        a = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        b = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertEqual(pk(a, b, **self.kwargs),
                         Decimal('0.2727272727272727272727272727'))
        self.assertEqual(pk(b, a, **self.kwargs),
                         Decimal('0.2727272727272727272727272727'))

    def test_all_kwargs_hyp_ref(self):
        '''
        Test whether a full miss and a translated boundary out of 4 produces
        0.273. 
        '''
        # pylint: disable=C0324,C0103
        metric_kwargs = dict(self.kwargs)
        metric_kwargs['hypothesis'] = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        metric_kwargs['reference'] = [1,1,1,1,1,2,3,3,4,4,4,4,4]
        self.assertEqual(pk(**metric_kwargs),
                         Decimal('0.2727272727272727272727272727'))

    def test_parts(self):
        '''
        Test parts.
        '''
        # pylint: disable=C0324,C0103
        a = [1,2,3,4,5,6,7,8,9,10,11,12,13]
        b = [1,1,1,1,2,2,2,2,3,3,3,3,3]
        metric_kwargs = dict(self.kwargs)
        metric_kwargs['return_parts'] = True
        self.assertEqual(pk(a, b, **metric_kwargs),
                         (7, 11))
    
    def test_format_exception(self):
        '''
        Test format exception.
        '''
        a = [2, 3, 6]
        b = [2, 2, 7]
        self.assertRaises(SegmentationMetricError, pk,
                          a, b, boundary_format=BoundaryFormat.sets)

    def test_mass_exception(self):
        '''
        Test length mismatch exception.
        '''
        a = [2, 2, 7]
        b = [2, 2, 8]
        self.assertRaises(SegmentationMetricError, pk, a, b)


class TestPairwisePkMeasure(TestCase):
    # pylint: disable=R0904,E1101,W0232
    '''
    Test pairwise Pk.
    '''
    
    def test_kazantseva2012_g5(self):
        '''
        Calculate permuted pairwise Pk on Group 5 from the dataset
        collected in [KazantsevaSzpakowicz2012]_.
        '''
        self.assertAlmostEquals(summarize(pk(KAZANTSEVA2012_G5)),
                         (Decimal('0.35530058282396693'),
                          Decimal('0.11001760846099215'),
                          Decimal('0.012103874171476172'),
                          Decimal('0.015879673965138168'),
                          48))
    
    def test_kazantseva2012_g2(self):
        '''
        Calculate mean permuted pairwise Pk on Group 2 from the dataset
        collected in [KazantsevaSzpakowicz2012]_.
        '''
        self.assertAlmostEquals(summarize(pk(KAZANTSEVA2012_G2)),
                         (Decimal('0.2882256923776327507173609771'),
                          Decimal('0.1454395656787966169084191445'),
                          Decimal('0.02115266726483699483402909754'),
                          Decimal('0.01327675514600517730547602481'),
                          120))
    
    def test_large_disagreement(self):
        '''
        Calculate mean permuted pairwise Pk on a theoretical dataset
        containing large disagreement.
        '''
        self.assertAlmostEquals(summarize(pk(LARGE_DISAGREEMENT)),
                         (1.0,
                          0.0,
                          0.0,
                          0.0,
                          8))
    
    def test_complete_agreement(self):
        '''
        Calculate mean permuted pairwise Pk on a theoretical dataset
        containing complete agreement.
        '''
        self.assertAlmostEquals(summarize(pk(COMPLETE_AGREEMENT)),
                         (0.0,
                          0.0,
                          0.0,
                          0.0,
                          48))
    
    def test_dataset_kwargs(self):
        '''
        Calculate mean permuted pairwise Pk on a theoretical dataset
        containing complete agreement.
        '''
        self.assertAlmostEquals(summarize(pk(dataset=COMPLETE_AGREEMENT)),
                         (0.0,
                          0.0,
                          0.0,
                          0.0,
                          48))

    def test_pk_datasets(self):
        '''
        Test pk upon two datasets.
        '''
        hypothesis = HYPOTHESIS_STARGAZER
        reference = HEARST_1997_STARGAZER
        value = pk(hypothesis, reference)
        self.assertAlmostEquals(float(value['stargazer,h1,1']), 0.26315789)
        self.assertAlmostEquals(float(value['stargazer,h2,1']), 0.36842105)
        self.assertAlmostEquals(float(value['stargazer,h1,2']), 0.42105263)
        self.assertAlmostEquals(float(value['stargazer,h2,2']), 0.42105263)

